<?
/* QM Promisance - Turn-based strategy game
 * Copyright (C) 2001-2011 QMT Productions
 *
 * $URL: https://qmpromisance.svn.sourceforge.net/svnroot/qmpromisance/promisance/tags/release-4_5/includes/database.php $
 * $Id: database.php 1848 2011-04-02 00:12:46Z Quietust $
 */

if (!defined('IN_GAME'))
	die("Access denied");

// Supported database types:
// 'mysql' - MySQL 4.1 or greater
// 'pgsql' - PostgreSQL 8.1 or greater
// 'sqlite' - SQLite 3 or greater (requires PHP 5.3 or greater)

function db_open ($type, $sock, $host, $port, $user, $pass, $name)
{
	if (!strlen($sock) && !strlen($host))
		die('Invalid database configuration - must specify either hostname or UNIX socket');
	if (strlen($sock) && strlen($host))
		die('Invalid database configuration - cannot specify both hostname and UNIX socket');

	try
	{
		switch ($type)
		{
		case 'mysql':
			$db = QM_PDO_MYSQL::open($sock, $host, $port, $user, $pass, $name);
			break;

		case 'pgsql':
			$db = QM_PDO_PGSQL::open($sock, $host, $port, $user, $pass, $name);
			break;

		case 'sqlite':
			$db = QM_PDO_SQLITE::open($sock, $host, $port, $user, $pass, $name);
			break;

		default:
			die('An unsupported database driver has been specified!');
			break;
		}
	}
	catch (PDOException $e)
	{
		return NULL;
	}

	return $db;
}

// Wrapper for PDO, adds query counting and some additional methods for common database-bound actions
abstract class QM_PDO extends PDO
{
	protected $queryCount;

	public function __construct($dsn, $username = '', $password = '', $driver_options = array())
	{
		parent::__construct($dsn, $username, $password, $driver_options);
		$this->setAttribute(PDO::ATTR_STATEMENT_CLASS, array('QM_PDOStatement', array($this, &$this->queryCount)));
		$this->setAttribute(PDO::ATTR_ERRMODE, PDO::ERRMODE_SILENT);
		$this->queryCount = 0;
	}

	// Retrieve current query count
	public function getQueryCount ()
	{
		return $this->queryCount;
	}

	// Placeholder for query logging
	public function query ($statement)
	{
		$this->queryCount++;
		return parent::query($statement);
	}

	// 1-liner for executing quick SQL queries, intended only for turns.php
	public function queryParam ($query, $parms = NULL)
	{
		$q = $this->prepare($query) or warning('Failed to prepare SQL query', 1);
		if ($parms)
			$q->bindAllValues($parms);
		$q->execute() or warning('Failed to execute SQL query', 1);
		return $q;
	}

	// evaluate an SQL query, return first cell of first row
	// useful for "SELECT COUNT(*) ..." queries
	public function queryCell ($query, $parms = NULL)
	{
		$q = $this->queryParam($query, $parms);
		return $q->fetchColumn();
	}

	// Clear the contents of a table and reset its auto-incrementing ID
	abstract public function clearTable ($table);

	// Retrieves a database's sequence, if it exists
	abstract public function getSequence ($table);

	// Modifies the query to limit rows returned
	abstract public function setLimit ($query, $rows, $offset = 0);

	// Lock a single entity
	abstract public function lockEntity ($table, $col, $type, $id, $owner);

	// Unlock a single entity
	abstract public function unlockEntity ($table, $col, $type, $id, $owner);

	// Lock all entities of a particular type
	// Cannot be used on per-table locks - instead, just use lockEntity
	abstract public function lockEntities ($table, $col, $type, $owner);

	// Unlock all entities of a particular type
	// Cannot be used on per-table locks - instead, just use unlockEntity
	abstract public function unlockEntities ($table, $col, $type, $owner);

	// Fetches all locks of a particular type
	abstract public function getLocks ($table, $col, $type);
}

// Subclass of QM_PDO for MySQL-specific behavior
class QM_PDO_MYSQL extends QM_PDO
{
	public static function open ($sock, $host, $port, $user, $pass, $name)
	{
		$dsn = 'mysql:';
		if ($sock)
			$dsn .= 'unix_socket='. $sock .';';
		elseif ($host)
		{
			$dsn .= 'host='. $host .';';
			if ($port)
				$dsn .= 'port='. $port .';';
		}
		$dsn .= 'dbname='. $name;

		return new QM_PDO_MYSQL($dsn, $user, $pass);
	}

	public function __construct($dsn, $username = '', $password = '', $driver_options = array())
	{
		parent::__construct($dsn, $username, $password, $driver_options);

		$this->exec("SET NAMES 'utf8'");
		$this->exec("SET SESSION sql_mode=''");
	}

	public function clearTable ($table)
	{
		if (!$this->query('TRUNCATE TABLE '. $table))
			return FALSE;
		return TRUE;
	}

	public function getSequence ($table)
	{
		return NULL;
	}

	public function setLimit ($query, $rows, $offset = 0)
	{
		if ($offset)
			$query .= ' LIMIT '. $offset .','. $rows;
		else	$query .= ' LIMIT '. $rows;
		return $query;
	}

	public function lockEntity ($table, $col, $type, $id, $owner)
	{
		$q = $this->prepare('INSERT INTO '. LOCK_TABLE .' (lock_type,lock_id,lock_owner) VALUES (?,?,?)');
		$q->bindIntValue(1, $type);
		$q->bindIntValue(2, $id);
		$q->bindIntValue(3, $owner);

		// failure to insert likely means there was a key violation, meaning it was already locked
		if (!$q->execute())
			return FALSE;
		return TRUE;
	}

	public function unlockEntity ($table, $col, $type, $id, $owner)
	{
		if ($owner)
		{
			$q = $this->prepare('DELETE FROM '. LOCK_TABLE .' WHERE lock_type = ? AND lock_id = ? AND lock_owner = ?');
			$q->bindIntValue(1, $type);
			$q->bindIntValue(2, $id);
			$q->bindIntValue(3, $owner);
		}
		else
		{
			$q = $this->prepare('DELETE FROM '. LOCK_TABLE .' WHERE lock_type = ? AND lock_id = ?');
			$q->bindIntValue(1, $type);
			$q->bindIntValue(2, $id);
		}

		if (!$q->execute())
			return FALSE;

		// can happen if a lock is forcibly released from admin/locks.php
		// this should still return TRUE and mark the entity as unlocked
		if ($q->rowCount() == 0)
			warning('Entity '. $id .' of type '. $type .' abruptly lost database lock', 2);

		return TRUE;
	}

	public function lockEntities ($table, $col, $type, $owner)
	{
		while (1)
		{
			$q = $this->prepare('INSERT IGNORE INTO '. LOCK_TABLE .' (lock_type,lock_id,lock_owner) SELECT ?,'. $col .',? FROM '. $table);
			$q->bindIntValue(1, $type);
			$q->bindIntValue(2, $owner);
			$q->execute() or warning('Failed to mass lock entities of type '. $type, 2);

			if ($this->queryCell('SELECT COUNT(*) FROM '. $table .' WHERE '. $col .' NOT IN (SELECT lock_id FROM '. LOCK_TABLE .' WHERE lock_type = ? AND lock_owner = ?)', array($type, $owner)) == 0)
				break;
			// delay 50-100ms and retry if any have not been locked by specified owner
			usleep(mt_rand(50000, 100000));
		}
	}

	public function unlockEntities ($table, $col, $type, $owner)
	{
		$q = $this->prepare('DELETE FROM '. LOCK_TABLE .' WHERE lock_type = ? AND lock_owner = ?');
		$q->bindIntValue(1, $type);
		$q->bindIntValue(2, $owner);
		$q->execute() or warning('Failed to mass unlock entities of type '. $type, 2);
	}

	public function getLocks ($table, $col, $type)
	{
		$q = $this->prepare('SELECT * FROM '. LOCK_TABLE .' WHERE lock_type = ?');
		$q->bindIntValue(1, $type);
		$q->execute() or warning('Failed to fetch locks of type '. $type, 1);
		return $q->fetchAll();
	}
}

// Subclass of QM_PDO for PostgreSQL-specific behavior
class QM_PDO_PGSQL extends QM_PDO
{
	public static function open ($sock, $host, $port, $user, $pass, $name)
	{
		$dsn = 'pgsql:';
		if ($sock)
			$dsn .= 'host='. $sock .' ';
		elseif ($host)
		{
			$dsn .= 'host='. $host .' ';
			if ($port)
				$dsn .= 'port='. $port .' ';
		}
		$dsn .= 'dbname='. $name;

		return new QM_PDO_PGSQL($dsn, $user, $pass);
	}

	public function clearTable ($table)
	{
		if (!$this->query('TRUNCATE TABLE '. $table))
			return FALSE;
		$seq = $this->getSequence($table);
		if ($seq)
		{
			if (!$this->query('ALTER SEQUENCE '. $seq .' RESTART WITH 1'))
				return FALSE;
		}
		return TRUE;
	}

	public function getSequence ($table)
	{
		global $sequences;
		// Not all tables have sequences associated with them
		if (!isset($sequences[$table]))
			return NULL;
		return $sequences[$table];
	}

	public function setLimit ($query, $rows, $offset = 0)
	{
		$query .= ' LIMIT '. $rows;
		if ($offset)
			$query .= ' OFFSET '. $offset;
		return $query;
	}

	public function lockEntity ($table, $col, $type, $id, $owner)
	{
		if ($type == ENT_VARS)
		{
			$q = $this->prepare('UPDATE '. VAR_TABLE .' SET v_value = ? WHERE v_name = ? AND v_value = ?');
			$q->bindStrValue(1, $owner);
			$q->bindStrValue(2, 'lock_id');
			$q->bindStrValue(3, 0);
		}
		else
		{
			$q = $this->prepare('UPDATE '. $table .' SET lock_id = ?, lock_date = NOW() WHERE '. $col .' = ? AND lock_id = 0');
			$q->bindIntValue(1, $owner);
			$q->bindIntValue(2, $id);
		}

		if (!$q->execute())
		{
			warning('Entity '. $id .' of type '. $type .' failed to lock', 2);
			return FALSE;
		}
		if ($q->rowCount() == 0)
			return FALSE;
		return TRUE;
	}

	public function unlockEntity ($table, $col, $type, $id, $owner)
	{
		if ($type == ENT_VARS)
		{
			if ($owner)
			{
				$q = $this->prepare('UPDATE '. VAR_TABLE .' SET v_value = ? WHERE v_name = ? AND v_value = ?');
				$q->bindIntValue(1, 0);
				$q->bindStrValue(2, 'lock_id');
				$q->bindStrValue(3, $owner);
			}
			else
			{
				$q = $this->prepare('UPDATE '. VAR_TABLE .' SET v_value = ? WHERE v_name = ?');
				$q->bindIntValue(1, 0);
				$q->bindStrValue(2, 'lock_id');
			}
		}
		else
		{
			if ($owner)
			{
				$q = $this->prepare('UPDATE '. $table .' SET lock_id = 0 WHERE '. $col .' = ? AND lock_id = ?');
				$q->bindIntValue(1, $id);
				$q->bindIntValue(2, $owner);
			}
			else
			{
				$q = $this->prepare('UPDATE '. $table .' SET lock_id = 0 WHERE '. $col .' = ?');
				$q->bindIntValue(1, $id);
			}
		}

		if (!$q->execute())
			return FALSE;
		if ($q->rowCount() == 0)
			return FALSE;
		return TRUE;
	}

	public function lockEntities ($table, $col, $type, $owner)
	{
		$q = $this->prepare('UPDATE '. $table .' SET lock_id = ?, lock_date = NOW() WHERE lock_id = 0');
		$q->bindIntValue(1, $owner);
		while (1)
		{
			$q->execute() or warning('Failed to mass lock entities of type '. $type, 2);

			if ($this->queryCell('SELECT COUNT(*) FROM '. $table .' WHERE lock_id != ?', array($owner)) == 0)
				break;
			// delay 50-100ms and retry if any have not been locked by specified owner
			usleep(mt_rand(50000, 100000));
		}
	}

	public function unlockEntities ($table, $col, $type, $owner)
	{
		$q = $this->prepare('UPDATE '. $table .' SET lock_id = 0 WHERE lock_id = ?');
		$q->bindIntValue(1, $owner);
		$q->execute() or warning('Failed to mass unlock entities of type '. $type, 2);
	}

	public function getLocks ($table, $col, $type)
	{
		if ($type == ENT_VARS)
		{
			$q = $this->prepare('SELECT v_value FROM '. VAR_TABLE .' WHERE v_name = ?');
			$q->bindStrValue(1, 'lock_id');
			$q->execute() or warning('Failed to fetch locks of type '. $type, 1);
			$id = $q->fetchColumn();
			if ($id > 0)
				return array(array('lock_id' => 1, 'lock_owner' => $id, 'lock_date' => 0));
			else	return array();
		}
		else
		{
			$q = $this->query('SELECT '. $col .' AS lock_id, lock_id AS lock_owner, lock_date FROM '. $table .' WHERE lock_id != 0');
			return $q->fetchAll();
		}
	}
}

// Subclass of QM_PDO for SQLite-specific behavior
class QM_PDO_SQLITE extends QM_PDO
{
	public static function open ($sock, $host, $port, $user, $pass, $name)
	{
		if ($host)
			return NULL;
		$dsn = 'sqlite:'. $sock;

		return new QM_PDO_SQLITE($dsn, $user, $pass);
	}
	public function __construct($dsn, $username = '', $password = '', $driver_options = array())
	{
		parent::__construct($dsn, $username, $password, $driver_options);

		$this->exec("PRAGMA encoding = 'UTF-8'");
		$this->exec("PRAGMA locking_mode = EXCLUSIVE");
		$this->exec("PRAGMA synchronous = OFF");
		$this->sqliteCreateFunction('LEAST', array('QM_PDO_SQLITE', 'sqlite_least'), 2);
	}

	public function clearTable ($table)
	{
		if (!$this->query('DELETE FROM '. $table))
			return FALSE;
		return TRUE;
	}

	public function getSequence ($table)
	{
		return NULL;
	}

	public function setLimit ($query, $rows, $offset = 0)
	{
		$query .= ' LIMIT '. $rows;
		if ($offset)
			$query .= ' OFFSET '. $offset;
		return $query;
	}

	public function lockEntity ($table, $col, $type, $id, $owner)
	{
		if ($type == ENT_VARS)
		{
			$q = $this->prepare('UPDATE '. VAR_TABLE .' SET v_value = ? WHERE v_name = ? AND v_value = 0');
			$q->bindIntValue(1, $owner);
			$q->bindStrValue(2, 'lock_id');
		}
		else
		{
			$q = $this->prepare('UPDATE '. $table .' SET lock_id = ?, lock_date = DATETIME(?) WHERE '. $col .' = ? AND lock_id = 0');
			$q->bindIntValue(1, $owner);
			$q->bindStrValue(2, 'now');
			$q->bindIntValue(3, $id);
		}

		if (!$q->execute())
		{
			warning('Entity '. $id .' of type '. $type .' failed to lock', 2);
			return FALSE;
		}
		if ($q->rowCount() == 0)
			return FALSE;
		return TRUE;
	}

	public function unlockEntity ($table, $col, $type, $id, $owner)
	{
		if ($type == ENT_VARS)
		{
			if ($owner)
			{
				$q = $this->prepare('UPDATE '. VAR_TABLE .' SET v_value = 0 WHERE v_name = ? AND v_value = ?');
				$q->bindStrValue(1, 'lock_id');
				$q->bindIntValue(2, $owner);
			}
			else
			{
				$q = $this->prepare('UPDATE '. VAR_TABLE .' SET v_value = 0 WHERE v_name = ?');
				$q->bindStrValue(1, 'lock_id');
			}
		}
		else
		{
			if ($owner)
			{
				$q = $this->prepare('UPDATE '. $table .' SET lock_id = 0 WHERE '. $col .' = ? AND lock_id = ?');
				$q->bindIntValue(1, $id);
				$q->bindIntValue(2, $owner);
			}
			else
			{
				$q = $this->prepare('UPDATE '. $table .' SET lock_id = 0 WHERE '. $col .' = ?');
				$q->bindIntValue(1, $id);
			}
		}

		if (!$q->execute())
			return FALSE;
		if ($q->rowCount() == 0)
			return FALSE;
		return TRUE;
	}

	public function lockEntities ($table, $col, $type, $owner)
	{
		$q = $this->prepare('UPDATE '. $table .' SET lock_id = ?, lock_date = DATETIME(?) WHERE lock_id = 0');
		$q->bindIntValue(1, $owner);
		$q->bindStrValue(2, 'now');
		while (1)
		{
			$q->execute() or warning('Failed to mass lock entities of type '. $type, 2);

			if ($this->queryCell('SELECT COUNT(*) FROM '. $table .' WHERE lock_id != ?', array($owner)) == 0)
				break;
			// delay 50-100ms and retry if any have not been locked by specified owner
			usleep(mt_rand(50000, 100000));
		}
	}

	public function unlockEntities ($table, $col, $type, $owner)
	{
		$q = $this->prepare('UPDATE '. $table .' SET lock_id = 0 WHERE lock_id = ?');
		$q->bindIntValue(1, $owner);
		$q->execute() or warning('Failed to mass unlock entities of type '. $type, 2);
	}

	public function getLocks ($table, $col, $type)
	{
		if ($type == ENT_VARS)
		{
			$q = $this->prepare('SELECT v_value FROM '. VAR_TABLE .' WHERE v_name = ?');
			$q->bindStrValue(1, 'lock_id');
			$q->execute() or warning('Failed to fetch locks of type '. $type, 1);
			$id = $q->fetchColumn();
			if ($id > 0)
				return array(array('lock_id' => 1, 'lock_owner' => $id, 'lock_date' => 0));
			else	return array();
		}
		else
		{
			$q = $this->query('SELECT '. $col .' AS lock_id, lock_id AS lock_owner, lock_date FROM '. $table .' WHERE lock_id != 0');
			return $q->fetchAll();
		}
	}

	// SQLite does not implement the LEAST() function (it uses MIN() instead), so it needs to be implemented here
	public static function sqlite_least ($num1, $num2)
	{
		return min($num1, $num2);
	}
}

// Wrapper for PDOStatement, adds query counting and some shortcuts for binding parameters
class QM_PDOStatement extends PDOStatement
{
	protected $dbh;
	protected $queryCount;

	protected function __construct ($dbh, &$queryCount)
	{
		$this->dbh = $dbh;
		$this->queryCount = &$queryCount;
	}

	// Placeholder for query logging
	public function execute ($input_parameters = NULL)
	{
		$this->queryCount++;
		return parent::execute($input_parameters);
	}

	// Shortcut for binding an integer parameter
	public function bindIntValue ($parm, $value)
	{
		return $this->bindValue($parm, $value, PDO::PARAM_INT);
	}

	// Shortcut for binding a string parameter
	public function bindStrValue ($parm, $value)
	{
		return $this->bindValue($parm, $value, PDO::PARAM_STR);
	}

	// Shortcut for binding a float parameter
	// Since PDO lacks a "float" type, bind as string for now
	public function bindFltValue ($parm, $value)
	{
		return $this->bindValue($parm, $value, PDO::PARAM_STR);
	}

	// Automatically bind all parameters based on type (for cases where the parameter count varies)
	public function bindAllValues ($parms)
	{
		foreach ($parms as $i => $val)
		{
			if (is_numeric($val))
			{
				if ($val - floor($val) > 0.00001)
					$this->bindFltValue($i + 1, $val);
				else	$this->bindIntValue($i + 1, $val);
			}
			else	$this->bindStrValue($i + 1, $val);
		}
	}
}
?>
